ARG_MARK = 1
WILDCARD_MARK = 2
FN_MARK = 3


class WordRouter:
    delimiter = "/"
    arg_char = ":"
    wildcard_char = "*"

    def __init__(self):
        self.routes = dict()

    def add(self, route, fn):
        wchar = self.wildcard_char
        achar = self.arg_char
        parent = self.routes
        for word in route.split(self.delimiter):
            firstchar = word[:1]
            if firstchar == achar:
                word = ARG_MARK

            elif firstchar == wchar:
                word = WILDCARD_MARK

            if word not in parent.keys():
                parent[word] = dict()

            parent = parent[word]

        parent[FN_MARK] = fn

    def dispatch(self, route):
        words = route.split(self.delimiter)
        route = self.routes
        route_keys = route.keys()
        route_args = []
        steps = 0
        for word in words:
            if word in route_keys:
                route = route[word]
                route_keys = route.keys()
                steps += 1

            elif ARG_MARK in route_keys:
                # Extract arguments
                route = route[ARG_MARK]
                route_keys = route.keys()
                route_args.append(word)
                steps += 1

            elif WILDCARD_MARK in route_keys:
                if WILDCARD_MARK in route:
                    route = route[WILDCARD_MARK]
                route_keys = (WILDCARD_MARK,)
                route_args.append(word)
                steps += 1

        if steps != len(words):
            return None, route_args

        return route.get(FN_MARK), route_args

    def __call__(self, route):
        fn, args = self.dispatch(route)
        if not fn:
            return
        return fn(*args)



import os

from os.path import isdir, join, relpath, pardir, exists
from time import strftime, gmtime
from mimetypes import guess_type
from .exceptions import HTTPNotFound, HTTPForbidden
from .constants import HTTP_DATETIME_FORMAT


class StaticHandlerMixin:
    def add_static(
        self,
        path: str,
        directory: str,
        default_document: str = "index.html",
        chunk_size: int = 0x4000,
    ):
        def get(*remaining_paths):
            response = self.response
            # response.charset = "utf-8"

            # Find the physical path of the given path parts
            physical_path = join(directory, *remaining_paths)

            # Check to do not access the parent directory of root and also we
            # are not listing directories here.
            if pardir in relpath(physical_path, directory):
                raise HTTPForbidden

            if isdir(physical_path):
                physical_path = join(physical_path, default_document)
                if not (default_document and exists(physical_path)):
                    raise HTTPNotFound()

            response.headers["content-type"] = (
                guess_type(physical_path)[0] or "application/octet-stream"
            )

            try:
                f = open(physical_path, mode="rb")
                stat = os.fstat(f.fileno())
                response.headers["content-length"] = str(stat[6])
                response.headers["last-modified"] = strftime(
                    HTTP_DATETIME_FORMAT, gmtime(stat.st_mtime)
                )

                with f:
                    while True:
                        r = f.read(chunk_size)
                        if not r:
                            break
                        yield r

            except OSError:
                raise HTTPNotFound

        self.route(path=path, formatter=lambda x, y: None)(get)
        self.route(path=f"{path}/*", formatter=lambda x, y: None)(get)



import cgi
import wsgiref.util as wsgiutil

try:
    import ujson as json
except ImportError:  # pragma: no cover
    import json

from http import cookies
from urllib.parse import parse_qs

from gongish.helpers import HeaderSet, LazyAttribute
from gongish.exceptions import HTTPBadRequest


def getcgifieldvalue(field):
    return (
        field.value
        if isinstance(field, cgi.MiniFieldStorage)
        or (isinstance(field, cgi.FieldStorage) and not field._binary_file)
        else field
    )


class RequestForm(dict):
    def __init__(self, environ, contenttype, contentlength):
        if contenttype == "application/json":
            if contentlength is None:
                raise HTTPBadRequest

            fp = environ["wsgi.input"]
            try:
                super().__init__(json.load(fp))
                return

            except (ValueError, TypeError):
                raise HTTPBadRequest("Cannot parse the request")

        if "QUERY_STRING" not in environ:
            environ["QUERY_STRING"] = ""

        try:
            storage = cgi.FieldStorage(
                fp=environ["wsgi.input"],
                environ=environ,
                strict_parsing=False,
                keep_blank_values=True,
            )

        except (TypeError, ValueError):
            raise HTTPBadRequest("Cannot parse the request")

        super().__init__()
        if storage.list is None or not len(storage.list):
            return

        for k in storage:
            v = storage[k]
            if isinstance(v, list):
                self[k] = [getcgifieldvalue(i) for i in v]
            else:
                self[k] = getcgifieldvalue(v)

    def __getattr__(self, key):
        if key not in self:
            return None
        return self[key]


class Request:
    def __init__(self, environ):
        self.environ = environ

    @LazyAttribute
    def verb(self):
        return self.environ["REQUEST_METHOD"].lower()

    @LazyAttribute
    def path(self):
        return self.environ["PATH_INFO"]

    @LazyAttribute
    def fullpath(self):
        """ Request full URI (includes query string) """
        return wsgiutil.request_uri(self.environ, include_query=True)

    @LazyAttribute
    def contentlength(self):
        v = self.environ.get("CONTENT_LENGTH")
        return None if not v or not v.strip() else int(v)

    @LazyAttribute
    def contenttype(self):
        contenttype = self.environ.get("CONTENT_TYPE")
        if contenttype:
            return contenttype.split(";")[0]
        return None

    @LazyAttribute
    def query(self):
        if "QUERY_STRING" not in self.environ:
            return {}

        return {
            k: v[0] if len(v) == 1 else v
            for k, v in parse_qs(
                self.environ["QUERY_STRING"],
                keep_blank_values=True,
                strict_parsing=False,
            ).items()
        }

    @LazyAttribute
    def form(self):
        return RequestForm(self.environ, self.contenttype, self.contentlength)

    @LazyAttribute
    def cookies(self):
        result = cookies.SimpleCookie()
        if "HTTP_COOKIE" in self.environ:
            result.load(self.environ["HTTP_COOKIE"])
        return result

    @LazyAttribute
    def scheme(self):
        """Request Scheme (http|https)
        """
        return wsgiutil.guess_scheme(self.environ)

    @LazyAttribute
    def headers(self):
        return HeaderSet().load_from_wsgi_environ(self.environ)



import importlib
from argparse import ArgumentParser
from wsgiref.simple_server import make_server


p = ArgumentParser(
    prog="gongish serve", description="Serve a WSGI application"
)
p.add_argument(
    "module",
    nargs="?",
    help="Module and application name (e.g: myapp:app)",
    type=str,
)
p.add_argument(
    "-b",
    "--bind",
    type=str,
    help="Bind address (default: localhost:8080)",
    default="localhost:8080",
)


def main(args):
    args = p.parse_args(args)
    module_name, module_attr = args.module.split(":")
    module = importlib.import_module(module_name)
    app = getattr(module, module_attr)

    bind = args.bind
    if bind.startswith(":"):
        host = "localhost"
        port = bind[1:]
    elif ":" in bind:
        host, port = bind.split(":")
    else:
        host = bind
        port = "8080"

    httpd = make_server(host, int(port), app)
    print(f"Serving http://{host}:{port}")
    httpd.serve_forever()



class LazyAttribute:  # pragma: nocover
    """
    ``LazyAttribute`` decorator is intended to promote a
    function call to object attribute. This means the
    function is called once and replaced with
    returned value.
    >>> class A:
    ...     def __init__(self):
    ...         self.counter = 0
    ...     @LazyAttribute
    ...     def count(self):
    ...         self.counter += 1
    ...         return self.counter
    >>> a = A()
    >>> a.count
    1
    >>> a.count
    1
    """

    __slots__ = ("f",)

    def __init__(self, f):
        self.f = f

    def __get__(self, obj, t=None):
        f = self.f
        if obj is None:
            return f
        val = f(obj)
        setattr(obj, f.__name__, val)
        return val



import types
from .helpers import HeaderSet


class Response:
    def __init__(self, app):
        self.app = app
        self.headers = HeaderSet()
        self.status = "200 OK"
        self.body = None
        self.type = None
        self.charset = None
        self.length = None
        self._firstchunk = None

    @property
    def contenttype(self):
        if not self.type:
            return None

        result = self.type
        if self.charset:
            result += f"; charset={self.charset}"

        return result

    def prepare_to_start(self):
        contenttype = self.contenttype
        if contenttype:
            self.headers["content-type"] = contenttype

        body = self.body
        if isinstance(body, types.GeneratorType):
            # Trying to get at least one element from the generator,
            # to force the method call till the second
            # `yield` statement
            self._firstchunk = next(body)
            if self.length is not None:
                self.headers["content-length"] = str(self.length)
        else:
            if body is None:
                body = []

            elif isinstance(body, (str, bytes)):
                body = [body]

            if self.charset:
                body = [i.encode(self.charset) for i in body]

            self.headers["content-length"] = str(
                sum(len(i) for i in body)
                if self.length is None
                else self.length
            )
            self.body = body

    def startstream(self):
        # encode if required
        if self.charset and not isinstance(self._firstchunk, bytes):
            yield self._firstchunk.encode(self.charset)
            for chunk in self.body:
                yield chunk.encode(self.charset)

        else:
            yield self._firstchunk
            for chunk in self.body:
                yield chunk

        self.app.on_end_response()  # hook

    def start(self):
        if self._firstchunk is not None:
            return self.startstream()

        self.app.on_end_response()  # hook
        return self.body



from collections import OrderedDict


class HeaderSet(OrderedDict):  # pragma: nocover
    def __init__(self, items=None):
        super().__init__()
        for i in items or []:
            self.add(i)

    def add(self, k, *args):
        if isinstance(k, str):
            values = []
            if ":" in k:
                k, v = k.split(":", 1)
                values += [i.strip() for i in v.split(";")]

        else:
            k, v = k
            values = [v]

        if args:
            values += args

        self[k] = "; ".join(values)

    def __getitem__(self, k):
        return super().__getitem__(k.lower())

    def __setitem__(self, k, v):
        return super().__setitem__(k.lower(), v)

    def __delitem__(self, k):
        return super().__delitem__(k)

    def __iter__(self):
        for k, v in self.items():
            yield f"{k}: {v}"

    def __str__(self):
        return "\r\n".join(self)

    def __iadd__(self, other):
        for i in other:
            self.add(i)

        return self

    def load_from_wsgi_environ(self, environ):
        for key, value in environ.items():
            if key.startswith("HTTP_") and key not in (
                "HTTP_CONTENT_TYPE",
                "HTTP_CONTENT_LENGTH",
            ):
                header_name = key[5:].replace("_", "-")
                self[header_name] = value

            elif key in ("CONTENT_TYPE", "CONTENT_LENGTH") and value:
                header_name = key.replace("_", "-")
                self[header_name] = value

        return self



import os

from pymlconf import Root


class ConfigurationMixin:
    def __init__(self):
        from gongish import __version__

        context = {"gongish_version": __version__}
        context.update(os.environ)
        data = {"debug": True}
        self.config = Root(data=data, context=context)

    def configure(self, filename: str):
        self.config.loadfile(filename)
